//////////////////////////////////////////////////////////////////////
// Title: TMX Loader
// Author: Phaelax
// Date: May 6, 2013
//
// Commands:
//    tmx_LoadMap(string filename)
//    tmx_FreeMap()
//    tmx_DrawGridTile(int tileIndex, int red, int green, int blue)
//    tmx_SetMapPosition(int x, int y)
//    tmx_SetMapVisible(boolean)
//    int = tmx_GetTileIndex(int layer, int worldX, int worldY)
//
//////////////////////////////////////////////////////////////////////



//////////////////////////////////////////////////
// Initializes the TMX engine variables
//////////////////////////////////////////////////
_tmx_init:
    #CONSTANT TMX_ORIENTATION_ORTHO = 1
    #CONSTANT TMX_ORIENTATION_ISO   = 2
    #CONSTANT TMX_ORIENTATION_STAG  = 3


    Type TMX_Map_Data_Properties
        mapWidth    as integer
        mapHeight   as integer
        tileWidth   as integer
        tileHeight  as integer
        orientation as integer
        tileWidth2  as integer
        tileHeight2 as integer
        layerTotal  as integer
        offsetX     as integer
        offsetY     as integer
    EndType



    Type TMX_Tile
        tilesetIndex as integer
        sprite       as integer
        gid          as integer
    EndType


    Type TMX_Tileset
        firstGid    as integer
        lastGid     as integer
        tileWidth   as integer
        tileHeight  as integer
        name        as string
        imgWidth    as integer
        imgHeight   as integer
        source      as string
        sprite      as integer
    EndType


    Global TMX_Map as TMX_Map_Data_Properties
    Global _TilesetCount = 0

    dim TMX_Tilesets[5] as TMX_Tileset
RETURN




//////////////////////////////////////////////////
// Returns the tile number under world coordinates
// [x,y] in relation to the specified layer number
//////////////////////////////////////////////////
function tmx_DrawGridTile(t, r, g, b)
    if TMX_Map.orientation = TMX_ORIENTATION_ORTHO then t = tmx_DrawOrthoGridTile(t, r, g, b)
    if TMX_Map.orientation = TMX_ORIENTATION_ISO   then t = tmx_DrawIsoGridTile(t, r, g, b)
    if TMX_Map.orientation = TMX_ORIENTATION_STAG  then t = tmx_DrawStagGridTile(t, r, g, b)
endfunction t



//////////////////////////////////////////////////
//
//////////////////////////////////////////////////
function tmx_DrawOrthoGridTile(t, r, g, b)
    t = t - 1
    ty = floor(t / TMX_Map.mapWidth)
    tx = t - ty*TMX_Map.mapWidth

    x = tx*TMX_Map.tileWidth  + TMX_Map.offsetX
    y = ty*TMX_Map.tileHeight + TMX_Map.offsetY

    drawLine(x, y, x+TMX_Map.tileWidth, y, r,g,b)
    drawLine(x, y+TMX_Map.tileHeight, x+TMX_Map.tileWidth, y+TMX_Map.tileHeight, r,g,b)
    drawLine(x, y, x, y+TMX_Map.tileHeight, r,g,b)
    drawLine(x+TMX_Map.tileWidth, y, x+TMX_Map.tileWidth, y+TMX_Map.tileHeight, r,g,b)
endfunction t



//////////////////////////////////////////////////
//
//////////////////////////////////////////////////
function tmx_DrawIsoGridTile(t, r, g, b)
    t = t - 1
    ty = floor(t / TMX_Map.mapWidth)
    tx = t - ty*TMX_Map.mapWidth

    tw = TMX_Map.tileWidth / 2
    th = TMX_Map.tileHeight / 2

    x = (tx-ty-1)*tw + TMX_Map.offsetX
    y = (tx+ty-1)*th + TMX_Map.offsetY

    drawLine(x, y+TMX_Map.tileHeight, x+tw, y+th, r,g,b)
    drawLine(x+TMX_Map.tileWidth, y+TMX_Map.tileHeight, x+tw, y+TMX_Map.tileHeight+th, r,g,b)
    drawLine(x+tw, y+th, x+TMX_Map.tileWidth, y+TMX_Map.tileHeight, r,g,b)
    drawLine(x, y+TMX_Map.tileHeight, x+tw, y+TMX_Map.tileHeight+th, r,g,b)
endfunction t



//////////////////////////////////////////////////
//
//////////////////////////////////////////////////
function tmx_DrawStagGridTile(t, r, g, b)
    t = t - 1
    ty = floor(t / TMX_Map.mapWidth)
    tx = t - ty*TMX_Map.mapWidth

    tw = TMX_Map.tileWidth / 2
    th = TMX_Map.tileHeight / 2

    if ty mod 2 = 1 then shiftX = tw
    x = tx*TMX_Map.tileWidth + shiftX + TMX_Map.offsetX
    y = ty*th + TMX_Map.offsetY

    drawLine(x, y+TMX_Map.tileHeight, x+tw, y+th, r,g,b)
    drawLine(x+TMX_Map.tileWidth, y+TMX_Map.tileHeight, x+tw, y+TMX_Map.tileHeight+th, r,g,b)
    drawLine(x+tw, y+th, x+TMX_Map.tileWidth, y+TMX_Map.tileHeight, r,g,b)
    drawLine(x, y+TMX_Map.tileHeight, x+tw, y+TMX_Map.tileHeight+th, r,g,b)
endfunction t



//////////////////////////////////////////////////
// Returns the tile number under world coordinates
// [x,y] in relation to the specified layer number
//////////////////////////////////////////////////
function tmx_GetTileIndex(layer, x, y)
    x = x - TMX_Map.offsetX
    y = y - TMX_Map.offsetY
    if TMX_Map.orientation = TMX_ORIENTATION_ORTHO then t = tmx_GetOrthoTile( x, y)
    if TMX_Map.orientation = TMX_ORIENTATION_ISO   then t = tmx_GetIsoTile(x, y)
    if TMX_Map.orientation = TMX_ORIENTATION_STAG  then t = tmx_GetStagTile(x, y)
    t = t + (layer-1) * TMX_Map.mapWidth*TMX_Map.mapHeight
endfunction t



//////////////////////////////////////////////////
//
//////////////////////////////////////////////////
function tmx_GetOrthoTile(x, y)
    tx = x / TMX_Map.tileWidth
    ty = y / TMX_Map.tileHeight
    t = tx + ty*TMX_Map.mapWidth + 1
endfunction t



//////////////////////////////////////////////////
//
//////////////////////////////////////////////////
function tmx_GetIsoTile(x, y)
    a = TMX_Map.tileWidth*y
    b = TMX_Map.tileHeight*x
    c = TMX_Map.tileWidth*TMX_Map.tileHeight
    tx = (a + b) / c
    ty = (a - b) / c
    t = tx + ty*TMX_Map.mapWidth + 1
endfunction t



//////////////////////////////////////////////////
//
//////////////////////////////////////////////////
function tmx_GetStagTile(x, y)
    th = TMX_Map.tileHeight/2

    dec y, th

    ratio# = TMX_Map.tileHeight / (tmx_Map.tileWidth+0.0)

    // Coordinates of grid-aligned tile
    tx = floor(x / TMX_Map.tileWidth)
    ty = floor(y / TMX_Map.tileHeight) * 2

    // relative position of the base square of the grid-aligned tile
    rx# = x - tx * TMX_Map.tileWidth
    ry# = y - (ty / 2) * TMX_Map.tileHeight

    if th - rx#*ratio# > ry#
        if ty mod 2
            dec ty
        else
            dec tx : dec ty
        endif
    endif

    if -th + rx#*ratio# > ry#
        if ty mod 2
            inc tx : dec ty
        else
            dec ty
        endif
    endif

    if th + rx#*ratio# < ry#
        if ty mod 2
            inc ty
        else
            dec tx : inc ty
        endif
    endif

    if th*3 - rx#*ratio# < ry#
        if ty mod 2
            inc tx : inc ty
        else
            inc ty
        endif
    endif

    t = tx + ty*TMX_Map.mapWidth + 1
endfunction t



//////////////////////////////////////////////////
// Position the map in world space
//////////////////////////////////////////////////
function tmx_SetMapPosition(ox, oy)
    if TMX_Map.orientation = TMX_ORIENTATION_ORTHO then tmx_PositionOrthoMap(ox, oy)
    if TMX_Map.orientation = TMX_ORIENTATION_ISO   then tmx_PositionIsoMap(ox, oy)
    if TMX_Map.orientation = TMX_ORIENTATION_STAG  then tmx_PositionStagMap(ox, oy)
    TMX_Map.offsetX = ox
    TMX_Map.offsetY = oy
endfunction



//////////////////////////////////////////////////
// Position the map in world space using an
// orthogonal layout
//////////////////////////////////////////////////
function tmx_PositionOrthoMap(ox, oy)
    for layer = 1 to TMX_Map.layerTotal
        for y = 0 to TMX_Map.mapHeight-1
            for x = 0 to TMX_Map.mapWidth-1
                t = x + y*TMX_Map.mapWidth + 1

                if TMX_Map[layer,t].tilesetIndex > 0
                    tx = x*TMX_Map.tileWidth
                    ty = y*TMX_Map.tileHeight
                    setSpritePosition(TMX_Map[layer, t].sprite, tx+ox, ty+oy)
                endif
            next x
        next y
    next layer
endfunction



//////////////////////////////////////////////////
// Position the map in world space using an
// isometric layout
//////////////////////////////////////////////////
function tmx_PositionIsoMap(ox, oy)
    tw = TMX_Map.tileWidth / 2
    th = TMX_Map.tileHeight / 2

    for layer = 1 to TMX_Map.layerTotal
        for y = 0 to TMX_Map.mapHeight-1
            for x = 0 to TMX_Map.mapWidth-1
                t = x + y*TMX_Map.mapWidth + 1

                if TMX_Map[layer,t].tilesetIndex > 0
                    tx = (x-y-1)*tw
                    ty = (x+y-1)*th
                    setSpritePosition(TMX_Map[layer, t].sprite, tx+ox, ty+oy)
                endif
            next x
        next y
    next layer
endfunction



//////////////////////////////////////////////////
// Position the map in world space using an
// isometric staggered layout
//////////////////////////////////////////////////
function tmx_PositionStagMap(ox, oy)
    tw = TMX_Map.tileWidth / 2
    th = TMX_Map.tileHeight / 2
    for layer = 1 to TMX_Map.layerTotal
        for y = 0 to TMX_Map.mapHeight-1
            shiftX = 0
            if y mod 2 = 1 then shiftX = tw
            for x = 0 to TMX_Map.mapWidth-1
                t = x + y*TMX_Map.mapWidth + 1

                if TMX_Map[layer,t].tilesetIndex > 0
                    tx = x*TMX_Map.tileWidth + shiftX
                    ty = y*th

                    setSpritePosition(TMX_Map[layer, t].sprite, tx+ox, ty+oy)
                endif
            next x
        next y
    next layer
endfunction



//////////////////////////////////////////////////
// Sets the map's visibility ON or OFF
//////////////////////////////////////////////////
function tmx_SetMapVisible(bool)
    for layer = 1 to TMX_Map.layerTotal
        for y = 0 to TMX_Map.mapHeight-1
            for x = 0 to TMX_Map.mapWidth-1
                t = x + y*TMX_Map.mapWidth + 1
                if TMX_Map[layer,t].tilesetIndex > 0 then setSpriteVisible(TMX_Map[layer, t].sprite, bool)
            next x
        next y
    next layer
endfunction



//////////////////////////////////////////////////
// Creates sprites and lays out the map layers
// for an orthogonal map
//////////////////////////////////////////////////
function tmx_BuildOrthogonalMap()
    for layer = 1 to TMX_Map.layerTotal
        for y = 0 to TMX_Map.mapHeight-1
            for x = 0 to TMX_Map.mapWidth-1
                t = x + y*TMX_Map.mapWidth + 1

                if TMX_Map[layer,t].tilesetIndex > 0
                    TMX_Map[layer, t].sprite = cloneSprite(TMX_Tilesets[TMX_Map[layer,t].tilesetIndex].sprite)
                    tx = x*TMX_Map.tileWidth
                    ty = y*TMX_Map.tileHeight
                    setSpritePosition(TMX_Map[layer, t].sprite, tx, ty)
                    setSpriteDepth(TMX_Map[layer, t].sprite, 10+j)
                    setSpriteFrame(TMX_Map[layer, t].sprite, TMX_Map[layer, t].gid)
                    setSpriteVisible(TMX_Map[layer, t].sprite, 1)
                endif
            next x
        next y
    next layer
endfunction



//////////////////////////////////////////////////
// Creates sprites and lays out the map layers
// for an orthogonal map
//////////////////////////////////////////////////
function tmx_BuildIsometricMap()
    tw = TMX_Map.tileWidth / 2
    th = TMX_Map.tileHeight / 2
    for layer = 1 to TMX_Map.layerTotal
        for y = 0 to TMX_Map.mapHeight-1
            for x = 0 to TMX_Map.mapWidth-1
                t = x + y*TMX_Map.mapWidth + 1

                if TMX_Map[layer,t].tilesetIndex > 0
                    TMX_Map[layer, t].sprite = cloneSprite(TMX_Tilesets[TMX_Map[layer,t].tilesetIndex].sprite)
                    tx = (x-y-1)*tw
                    ty = (x+y-1)*th

                    setSpritePosition(TMX_Map[layer, t].sprite, tx, ty)
                    setSpriteDepth(TMX_Map[layer, t].sprite, 10+j)
                    setSpriteFrame(TMX_Map[layer, t].sprite, TMX_Map[layer, t].gid)
                    setSpriteVisible(TMX_Map[layer, t].sprite, 1)
                endif
            next x
        next y
    next layer
endfunction



//////////////////////////////////////////////////
// Creates sprites and lays out the map layers
// for an orthogonal map
//////////////////////////////////////////////////
function tmx_BuildStaggeredMap()
    tw = TMX_Map.tileWidth / 2
    th = TMX_Map.tileHeight / 2
    for layer = 1 to TMX_Map.layerTotal
        for y = 0 to TMX_Map.mapHeight-1
            shiftX = 0
            if y mod 2 = 1 then shiftX = tw
            for x = 0 to TMX_Map.mapWidth-1
                t = x + y*TMX_Map.mapWidth + 1

                if TMX_Map[layer,t].tilesetIndex > 0
                    TMX_Map[layer, t].sprite = cloneSprite(TMX_Tilesets[TMX_Map[layer,t].tilesetIndex].sprite)
                    tx = x*TMX_Map.tileWidth + shiftX
                    ty = y*th

                    setSpritePosition(TMX_Map[layer, t].sprite, tx, ty)
                    setSpriteDepth(TMX_Map[layer, t].sprite, 10+j)
                    setSpriteFrame(TMX_Map[layer, t].sprite, TMX_Map[layer, t].gid)
                    setSpriteVisible(TMX_Map[layer, t].sprite, 1)
                endif
            next x
        next y
    next layer
endfunction



//////////////////////////////////////////////////
// Finds which tileset this gid belongs to
//////////////////////////////////////////////////
function tmx_GetTilesetIndex(gid)
    for i = 1 to 5
        if gid >= TMX_Tilesets[i].firstGid and gid <= TMX_Tilesets[i].lastGid then exitfunction i
    next i
endfunction 0



//////////////////////////////////////////////////
// Load an XML-based TMX map file
//////////////////////////////////////////////////
function tmx_LoadMap(file$)
    F = openToRead(file$)

    layerId = 0
    repeat
        s$ = xml_StripLeadingSpaces$(readline(F))
        tag$ = xml_GetTagName$(s$)


        if tag$ = "map"
            attributes$  = xml_GetUnparsedAttributeString$(s$, tag$)
            orientation$ = xml_GetAttribute$(attributes$, "orientation")
            TMX_Map.mapWidth    = val(xml_GetAttribute$(attributes$, "width"))
            TMX_Map.mapHeight   = val(xml_GetAttribute$(attributes$, "height"))
            TMX_Map.tileWidth   = val(xml_GetAttribute$(attributes$, "tilewidth"))
            TMX_Map.tileHeight  = val(xml_GetAttribute$(attributes$, "tileheight"))

            if orientation$ = "orthogonal" then TMX_Map.orientation = TMX_ORIENTATION_ORTHO
            if orientation$ = "isometric"  then TMX_Map.orientation = TMX_ORIENTATION_ISO
            if orientation$ = "staggered"  then TMX_Map.orientation = TMX_ORIENTATION_STAG

            // Create a map with a max of 5 layers
            dim TMX_Map[5, TMX_Map.mapWidth*TMX_Map.mapHeight] as TMX_Tile
        endif


        // New tileset detected
        if tag$ = "tileset"
            attributes$  = xml_GetUnparsedAttributeString$(s$, tag$)
            inc _TilesetCount
            TMX_Tilesets[_TilesetCount].firstGid   = val(xml_GetAttribute$(attributes$, "firstgid"))
            TMX_Tilesets[_TilesetCount].tileWidth  = val(xml_GetAttribute$(attributes$, "tilewidth"))
            TMX_Tilesets[_TilesetCount].tileHeight = val(xml_GetAttribute$(attributes$, "tileheight"))
            TMX_Tilesets[_TilesetCount].name       = xml_GetAttribute$(attributes$, "name")

        endif

        // Image detected, load it and set its animation to make up the tiles
        if tag$ = "image"
            attributes$  = xml_GetUnparsedAttributeString$(s$, tag$)
            TMX_Tilesets[_TilesetCount].source = xml_GetAttribute$(attributes$, "source")
            imgWidth   = val(xml_GetAttribute$(attributes$, "width"))
            imgHeight  = val(xml_GetAttribute$(attributes$, "height"))
            TMX_Tilesets[_TilesetCount].imgWidth = imgWidth
            TMX_Tilesets[_TilesetCount].imgHeight = imgHeight
            TMX_Tilesets[_TilesetCount].lastGid = floor(imgWidth / (TMX_Tilesets[_TilesetCount].tileWidth+0.0)) * floor(imgHeight / (TMX_Tilesets[_TilesetCount].tileHeight+0.0)) + TMX_Tilesets[_TilesetCount].firstGid - 1

            TMX_Tilesets[_TilesetCount].sprite = createSprite(loadImage(TMX_Tilesets[_TilesetCount].source))
            tileCount = (imgWidth / TMX_Tilesets[_TilesetCount].tileWidth) * (imgHeight / TMX_Tilesets[_TilesetCount].tileHeight)
            setSpriteAnimation(TMX_Tilesets[_TilesetCount].sprite, TMX_Tilesets[_TilesetCount].tileWidth, TMX_Tilesets[_TilesetCount].tileHeight, tileCount)
            setSpriteVisible(TMX_Tilesets[_TilesetCount].sprite, 0)
        endif


        // New layer detected
        if tag$ = "layer"
            inc layerId
            tileNumber = 0
        endif


        // Add tile to map's current layer
        if tag$ = "tile" and layerId > 0
            inc tileNumber
            attributes$ = xml_GetUnparsedAttributeString$(s$, tag$)
            gid = val(xml_GetAttribute$(attributes$, "gid"))
            tsi  = tmx_GetTilesetIndex(gid)
            TMX_Map[layerId, tileNumber].tilesetIndex = tsi
            TMX_Map[layerId, tileNumber].gid = gid - TMX_Tilesets[tsi].firstGid + 1
        endif


    until fileEOF(F) = 1
    closeFile(F)

    TMX_Map.layerTotal = layerId

    if TMX_Map.orientation = TMX_ORIENTATION_ORTHO then tmx_BuildOrthogonalMap()
    if TMX_Map.orientation = TMX_ORIENTATION_ISO   then tmx_BuildIsometricMap()
    if TMX_Map.orientation = TMX_ORIENTATION_STAG  then tmx_BuildStaggeredMap()

endfunction



//////////////////////////////////////////////////
// Returns the value of an attribute found in a
// string created by getUnparsedAttributeString$()
//////////////////////////////////////////////////
function tmx_FreeMap()
    count = TMX_Map.mapWidth * TMX_Map.mapHeight
    for layer = 1 to TMX_Map.layerTotal
        for i = 1 to count
            deleteSprite(TMX_Map[layer, i].sprite)
        next i
    next layer

    for i = 1 to _TilesetCount
        deleteSprite(TMX_Tilesets[i].sprite)
    next i


    _TilesetCount = 0
    undim dim TMX_Map[]
endfunction



//////////////////////////////////////////////////
// Returns the value of an attribute found in a
// string created by getUnparsedAttributeString$()
//////////////////////////////////////////////////
function xml_GetAttribute$(attributes$, attname$)
    attLength = len(attname$)

    L = len(attributes$)
    for i = 1 to L - attLength
        if mid(attributes$, i, attLength) = attname$ then exit
    next i
    v = i+attLength+1
    temp$ = str(v)


    opening = 1
    for i = v to L
        if mid(attributes$, i, 1) = chr(34)
            if open = 1
                closing = i-1
                exit
            endif
            if open = 0
                open = 1
                opening = i+1
            endif
        endif
    next i
    value$ = mid(attributes$, opening, closing-opening+1)

endfunction value$




//////////////////////////////////////////////////
// Returns a string stripped of its tag
//////////////////////////////////////////////////
function xml_GetUnparsedAttributeString$(text$, tag$)
    text$ = mid(text$, 3+len(tag$), -1)
endfunction text$




//////////////////////////////////////////////////
// Returns first tag name found in string
//////////////////////////////////////////////////
function xml_GetTagName$(text$)
    L = len(text$)
    for i = 2 to L
        c = asc(mid(text$, i, 1))
        if c = 62 or c = 32 then exit
    next i
    text$ = lower(mid(text$, 2, i-2))
endfunction text$




//////////////////////////////////////////////////
// Returns a string stripped of leading spaces
//////////////////////////////////////////////////
function xml_StripLeadingSpaces$(text$)
    s = 1
    for i = 1 to len(text$)
        if mid(text$, i, 1) <> " "
            s = i
            exit
        endif
    next i
    text$ = mid(text$, s, -1)
endfunction text$

